EPSILON = 0.0001

SEGMENT_NO_INTERSECT = 0
SEGMENT_INTERSECT = 1
SEGMENT_COLLINEAR = 2

def sign(number: float):
    if number < -EPSILON: return -1
    if number > EPSILON: return 1
    return 0

def dot(v, w):
    return v[0]*w[0] + v[1]*w[1]

def cos_angle(v, w):
    return dot(v, w)/((distance_sq((0, 0), v)*distance_sq((0, 0), w))**0.5)

def sin_angle(v, w):
    return signed_area2((0, 0), v, w)/((distance_sq((0, 0), v)*distance_sq((0, 0), w))**0.5)

def signed_area2(a, b, c):
    return (a[0]*b[1])+(a[1]*c[0])+(b[0]*c[1])-(c[0]*b[1])-(b[0]*a[1])-(a[0]*c[1]);

def middle(a, b):
    return ((a[0]+b[0])/2, (a[1]+b[1])/2)

def point_equals(p1, p2):
    return sign(p1[0]-p2[0]) == 0 and sign(p1[1]-p2[1]) == 0

def distance_sq(a, b):
    return ((a[0] - b[0]) ** 2) + ((a[1] - b[1]) ** 2 )

def determinant(a, b, c, d):
    return a * d - b * c

def point_position_segment(point, s1, s2):
    return sign(signed_area2(s1, s2, point))

def point_in_segment(point, s1, s2):
    if point_position_segment(point, s1, s2) == 0:
        v1 = (s1[0]-point[0], s1[1]-point[1])
        v2 = (s2[0]-point[0], s2[1]-point[1])

        if dot(v1, v2)<=0:
            return True

    return False


def point_in_triangle(point, t1, t2, t3):
    sa = sign(signed_area2(t1, t2, point))
    sb = sign(signed_area2(t2, t3, point))
    sc = sign(signed_area2(t3, t1, point))

    if sa and sb and sc:
        #None of the signed areas is 0, it's either inside or outside (no border)
        if sa==sb and sb==sc:
            if sa == 1: return 2
            if sa == -1: return -2
        return 0

    #It's on the border or outside
    if (not sa and point_in_segment(point, t1, t2)) or (not sb and point_in_segment(point, t2, t3)) or (not sc and point_in_segment(point, t3, t1)):
        return sign(signed_area2(t1, t2, t3))

    return 0

def intersection_point(s1, s2, t1, t2):
    a1 = s2[1] - s1[1]
    a2 = s1[0] - s2[0]
    a3 = t2[1] - t1[1]
    a4 = t1[0] - t2[0]

    denom = determinant(a1,a2,a3,a4)

    if sign(denom) == 0:
        return None

    #Compute X
    a1 = s1[0]*(s2[1]-s1[1]) - s1[1]*(s2[0]-s1[0]);
    a3 = t1[0]*(t2[1]-t1[1]) - t1[1]*(t2[0]-t1[0]);

    x = determinant(a1,a2,a3,a4) / denom;

    #Compute Y
    a1 = s2[1] - s1[1];
    a2 = s1[0]*(s2[1]-s1[1]) - s1[1]*(s2[0]-s1[0]);
    a3 = t2[1] - t1[1];
    a4 = t1[0]*(t2[1]-t1[1]) - t1[1]*(t2[0]-t1[0]);

    y = determinant(a1,a2,a3,a4) / denom;

    return (x, y);

def segment_position_segment(s1, s2, t1, t2):
    """
    Returns:
        - SEGMENT_INTERSECT if intersection_point() can be applied
        - SEGMENT_NO_INTERSECT if segments do not intersect.
        - SEGMENT_COLLINEAR if they are collinear (sharing a vertex or a sub-segment)
    """
    a = point_position_segment(s1, t1, t2)
    b = point_position_segment(s2, t1, t2)

    if a == b and a != 0:
        return SEGMENT_NO_INTERSECT

    c = point_position_segment(t1, s1, s2)
    d = point_position_segment(t2, s1, s2)

    if c == d and c != 0:
        return SEGMENT_NO_INTERSECT

    if not (a and b and c and d):
        return SEGMENT_COLLINEAR

    if a == -b and c == -d:
        return SEGMENT_INTERSECT

    if a == 0 and point_in_segment(s1, t1, t2):
        return SEGMENT_INTERSECT

    if b == 0 and point_in_segment(s2, t1, t2):
        return SEGMENT_INTERSECT

    if c == 0 and point_in_segment(t1, s1, s2):
        return SEGMENT_INTERSECT

    if d == 0 and point_in_segment(t2, s1, s2):
        return SEGMENT_INTERSECT

    return SEGMENT_NO_INTERSECT

def _sort_points(points, keyf):
    if len(points) < 3:
        return points
    else:
        return sorted(points, key=keyf)

def sort_points_distance(origin, points):
    key = lambda x: distance_sq(origin, x)
    return _sort_points(points, keyf=key)
